import random
from django.db import models
from datetime import timedelta
from django.conf import settings
from django.utils import timezone
from django.core.exceptions import ValidationError
from phonenumber_field.modelfields import PhoneNumberField

from .app_settings import app_settings


adapter = app_settings.ADAPTER()


class Account(models.Model):
    user = models.OneToOneField(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)

    is_verified = models.BooleanField(default=False)
    phone_verified = models.BooleanField(default=False)
    email_verified = models.BooleanField(default=False)

    verified_at = models.DateTimeField(null=True, blank=True)
    phone_verified_at = models.DateTimeField(null=True, blank=True)
    email_verified_at = models.DateTimeField(null=True, blank=True)

    def save(self, *args, **kwargs):
        if self.phone_verified or self.email_verified and not self.is_verified:
            self.is_verified = True

        if self.is_verified and not self.verified_at:
            self.verified_at = timezone.now()

        if self.phone_verified and not self.phone_verified_at:
            self.phone_verified_at = timezone.now()

        if self.email_verified and not self.email_verified_at:
            self.email_verified_at = timezone.now()

        return super().save(*args, **kwargs)


class TOTP(models.Model):
    PURPOSE_PASSWORD_RESET = "PasswordReset"
    PURPOSE_EMAIL_VERIFICATION = "EmailVerification"
    PURPOSE_PHONE_VERIFICATION = "PhoneVerification"
    PURPOSE_ACCOUNT_VERIFICATION = "AccountVerification"

    PURPOSE_CHOICES = [
        (PURPOSE_PASSWORD_RESET, "Password Reset"),
        (PURPOSE_EMAIL_VERIFICATION, "Email Verification"),
        (PURPOSE_PHONE_VERIFICATION, "Phone Verification"),
        (PURPOSE_ACCOUNT_VERIFICATION, "Account Verification"),
    ]

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
    )

    otp = models.IntegerField()
    is_valid = models.BooleanField(default=True)
    invalidated_at = models.DateTimeField(null=True, blank=True)
    purpose = models.CharField(max_length=100, choices=PURPOSE_CHOICES)
    expiration_time = models.DateTimeField()
    created_at = models.DateTimeField(auto_now_add=True)

    @property
    def is_expired(self):
        return timezone.now() > self.expiration_time

    def clean(self):
        if self.pk:
            original_state = TOTP.objects.get(pk=self.pk)
            if original_state.is_valid is False:
                raise ValidationError(
                    "is_valid cannot be reset to True once set to False."
                )
            if original_state.otp != self.otp:
                raise ValidationError("otp cannot be updated.")

        return super().clean()

    def save(self, *args, **kwargs):
        self.clean()

        # set `otp``
        otp_len = app_settings.OTP_LENGTH
        first_digit = str(random.randint(1, 9))
        digits = "".join([str(random.randint(0, 9)) for _ in range(otp_len - 1)])
        self.otp = int(first_digit + digits)

        # set `invalidated_at``
        if not self.is_valid and not self.invalidated_at:
            self.invalidated_at = timezone.now()

        # set `expiration_time`
        if not self.expiration_time:
            if self.purpose == self.PURPOSE_PASSWORD_RESET:
                self.expiration_time = timezone.now() + timedelta(
                    seconds=app_settings.PASSWORD_RESET_OTP_EXPIRY_TIME
                )
            else:
                self.expiration_time = timezone.now() + timedelta(
                    seconds=app_settings.OTP_EXPIRY_TIME
                )

        super().save(*args, **kwargs)

class TOTPMetadata(models.Model):
    totp = models.OneToOneField(TOTP, on_delete=models.CASCADE, related_name='metadata')
    new_phone = PhoneNumberField(null=True)
    new_email = models.EmailField(null=True)


# Generated by Django 5.0.7 on 2024-08-06 11:49

import django.db.models.deletion
import phonenumber_field.modelfields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('otp_rest_auth', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='TOTPMetadata',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('new_phone', phonenumber_field.modelfields.PhoneNumberField(max_length=128, null=True, region=None)),
                ('new_email', models.EmailField(max_length=254, null=True)),
                ('totp', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='metadata', to='otp_rest_auth.totp')),
            ],
        ),
    ]



from django.utils import timezone

from . import signals
from .models import TOTP
from .app_settings import app_settings

adapter = app_settings.ADAPTER()


def validate_otp(otp: int, purpose: str):
    totp = TOTP.objects.filter(otp=otp, purpose=purpose).first()
    if not totp:
        return False, None
    if totp.is_expired or not totp.is_valid:
        return False, None

    return True, totp


def verify_otp(otp: int, purpose: str):
    is_valid, totp = validate_otp(otp, purpose)
    if not is_valid:
        return False, None

    # Confirm OTP
    totp.is_valid = False
    totp.invalidated_at = timezone.now()
    totp.save()
    return True, totp


def send_verification_otp(totp: TOTP, request=None, signup=False):
    def signal_email_confirmation_sent():
        signals.email_confirmation_sent.send(
            sender=send_verification_otp,
            request=request,
            signup=signup,
        )

    def signal_phone_confirmation_sent():
        signals.phone_confirmation_sent.send(
            sender=send_verification_otp,
            request=request,
            signup=signup,
        )

    if totp.purpose == TOTP.PURPOSE_ACCOUNT_VERIFICATION:
        adapter.send_otp_to_user_email(totp)
        adapter.send_otp_to_user_phone(totp)

        signal_email_confirmation_sent()
        signal_phone_confirmation_sent()

    elif totp.purpose == TOTP.PURPOSE_EMAIL_VERIFICATION:
        adapter.send_otp_to_user_email(totp)
        signal_email_confirmation_sent()

    elif totp.purpose == TOTP.PURPOSE_PHONE_VERIFICATION:
        adapter.send_otp_to_user_phone(totp)
        signal_phone_confirmation_sent()

    elif totp.purpose == TOTP.PURPOSE_PASSWORD_RESET:
        medium = app_settings.PASSWORD_RESET_OTP_RECIPIENTS
        if "phone" in medium:
            adapter.send_otp_to_user_phone(totp)
        if "email" in medium:
            adapter.send_otp_to_user_email(totp)



from django.contrib.auth import get_user_model
from django.contrib.auth.forms import SetPasswordForm
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APIClient, APITestCase
from unittest.mock import patch, MagicMock
from otp_rest_auth.models import TOTP
from otp_rest_auth import app_settings
from otp_rest_auth.otp_ops import verify_otp, validate_otp

User = get_user_model()


class PasswordResetConfirmViewTests(APITestCase):
    def setUp(self):
        self.client = APIClient()
        self.user = User.objects.create_user(
            username="testuser",
            password="old_password",
            email="testuser@example.com",
            phone="+2348145640915",
        )
        self.totp = TOTP.objects.create(
            user=self.user, purpose=TOTP.PURPOSE_PASSWORD_RESET
        )
        self.url = reverse("otp_rest_password_reset_confirm")

    @patch("otp_rest_auth.serializers.verify_otp")
    @patch("otp_rest_auth.serializers.validate_otp")
    @patch("django.contrib.auth.forms.SetPasswordForm.is_valid", return_value=True)
    @patch("django.contrib.auth.forms.SetPasswordForm.save")
    def test_password_reset_confirm_success(
        self, mock_save, mock_is_valid, mock_validate_otp, mock_verify_otp
    ):
        mock_validate_otp.return_value = (True, self.totp)
        data = {
            "otp": self.totp.otp,
            "new_password1": "new_secure_password",
            "new_password2": "new_secure_password",
        }
        response = self.client.post(self.url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data["detail"], "Password has been reset with the new password."
        )
        mock_validate_otp.assert_called_once_with(
            data["otp"], TOTP.PURPOSE_PASSWORD_RESET
        )
        mock_verify_otp.assert_called_once_with(
            self.totp.otp, TOTP.PURPOSE_PASSWORD_RESET
        )
        mock_save.assert_called_once()

    @patch("otp_rest_auth.serializers.validate_otp")
    def test_password_reset_confirm_invalid_otp(self, mock_validate_otp):
        mock_validate_otp.side_effect = TOTP.DoesNotExist
        data = {
            "otp": "000000",
            "new_password1": "new_secure_password",
            "new_password2": "new_secure_password",
        }
        response = self.client.post(self.url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn("otp", response.data)

    def test_password_reset_confirm_invalid_passwords(self):
        data = {
            "otp": self.totp.otp,
            "new_password1": "short",
            "new_password2": "short",
        }

        response = self.client.post(self.url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn("new_password2", response.data)



from django.test import TestCase, RequestFactory
from rest_framework.test import APIClient
from rest_framework import status
from unittest.mock import patch
from django.contrib.auth import get_user_model
from django.test.utils import override_settings
from django.utils.translation import gettext_lazy as _

from otp_rest_auth.views import RegisterView
from otp_rest_auth.app_settings import app_settings

User = get_user_model()


class RegisterViewTests(TestCase):
    def setUp(self):
        self.client = APIClient()
        self.factory = RequestFactory()

    @patch("otp_rest_auth.otp_ops.send_verification_otp")
    @override_settings(
        OTP_REST_AUTH={"VERIFICATION_METHOD": app_settings.AccountVerificationMethod.NONE}
    )
    def test_create_user_without_verification(self, mock_send_verification_otp):
        """
        Test creating a user without verification disabled.
        """
        request_data = {
            "username": "testuser",
            "email": "test@example.com",
            "phone": "+2348145640915",
            "password1": "testPassword123!",
            "password2": "testPassword123!",
        }
        request = self.factory.post("/api/register/", request_data)

        view = RegisterView.as_view()
        response = view(request)

        self.assertFalse(mock_send_verification_otp.called)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertIn("user", response.data)
        self.assertIn("access", response.data)
        self.assertIn("refresh", response.data)

    @patch("otp_rest_auth.otp_ops.send_verification_otp")
    @patch("otp_rest_auth.models.TOTP.objects.create")
    @override_settings(
        OTP_REST_AUTH={"VERIFICATION_METHOD": app_settings.AccountVerificationMethod.EMAIL}
    )
    def test_create_user_with_verification(
        self, mock_create_totp, mock_send_verification_otp
    ):
        """
        Test creating a user with verification enabled.
        """
        request_data = {
            "username": "testuser",
            "email": "test@example.com",
            "phone": "+2348145640915",
            "password1": "testPassword123!",
            "password2": "testPassword123!",
        }
        request = self.factory.post("/api/register/", request_data)

        view = RegisterView.as_view()
        response = view(request)

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertIn("detail", response.data)
        self.assertEqual(response.data["detail"], _("Verification OTP sent."))

        # self.assertTrue(mock_create_totp.called)
        # self.assertTrue(mock_send_verification_otp.called)

    def test_invalid_data(self):
        """
        Test creating a user with invalid data.
        """
        request_data = {
            "username": "testuser",
            "email": "test@example.com",
            # Missing password fields
        }
        request = self.factory.post("/api/register/", request_data)

        view = RegisterView.as_view()
        response = view(request)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn("password1", response.data)
        self.assertIn("password2", response.data)



VERSION = (0, 2, 2, "final", 0)

__title__ = "otp-rest-auth"
__version_info__ = VERSION
__version__ = ".".join(map(str, VERSION[:3])) + (
    "-{}{}".format(VERSION[3], VERSION[4] or "") if VERSION[3] != "final" else ""
)
__author__ = "Leptons Multiconcept"
__license__ = "MIT"
__copyright__ = "Copyright (c) 2024 Leptons-Multiconcept"



from django.test import TestCase, RequestFactory
from django.urls import reverse
from django.contrib.auth import get_user_model
from rest_framework import status
from rest_framework.test import APIClient
from unittest.mock import patch
from unittest import skip
from django.test.utils import override_settings

from otp_rest_auth.models import Account
from otp_rest_auth.views import LoginView
from otp_rest_auth.app_settings import app_settings


User = get_user_model()


class LoginViewTests(TestCase):
    def setUp(self):
        self.factory = RequestFactory()
        self.view = LoginView.as_view()
        self.url = reverse("otp_rest_login")
        self.client = APIClient()
        self.user = User.objects.create_user(
            username="testuser",
            email="test@example.com",
            phone="+2348145640915",
            password="password",
        )

        self.account = Account.objects.get(user=self.user)
        self.account.is_verified = True
        self.account.save()

    @override_settings(
        OTP_REST_AUTH={
            "VERIFICATION_METHOD": app_settings.AccountVerificationMethod.ACCOUNT,
        }
    )
    def test_post_valid_credentials(self):
        data = {"username": "testuser", "password": "password"}
        response = self.client.post(self.url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn("access", response.data)
        self.assertIn("refresh", response.data)

    def test_post_invalid_credentials(self):
        data = {"username": "testuser", "password": "wrongpassword"}
        response = self.client.post(self.url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn("non_field_errors", response.data)

    @override_settings(
        OTP_REST_AUTH={
            "VERIFICATION_REQUIRED": True,
            "VERIFICATION_METHOD": app_settings.AccountVerificationMethod.ACCOUNT,
        }
    )
    @patch("otp_rest_auth.views.get_login_response_data")
    @patch("otp_rest_auth.views.set_jwt_cookies")
    def test_jwt_cookies_set(self, mock_set_jwt_cookies, mock_get_login_response_data):
        mock_get_login_response_data.return_value = {
            "user": {"username": "testuser"},
            "access": "access_token",
            "refresh": "refresh_token",
        }

        data = {"username": "testuser", "password": "password"}
        request = self.factory.post(self.url, data, format="json")
        response = self.view(request)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        mock_set_jwt_cookies.assert_called_once_with(
            response, "access_token", "refresh_token"
        )

    @override_settings(
        OTP_REST_AUTH={
            "VERIFICATION_REQUIRED": True,
            "VERIFICATION_METHOD": app_settings.AccountVerificationMethod.ACCOUNT,
        }
    )
    def test_phone_login_with_account_verification_type(self):
        data = {"phone": "+2348145640915", "password": "password"}
        response = self.client.post(self.url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn("access", response.data)
        self.assertIn("refresh", response.data)

    @override_settings(
        OTP_REST_AUTH={
            "VERIFICATION_REQUIRED": True,
            "VERIFICATION_METHOD": app_settings.AccountVerificationMethod.ACCOUNT,
        }
    )
    def test_email_login_with_account_verification_type(self):
        data = {"email": self.user.email, "password": "password"}
        response = self.client.post(self.url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn("access", response.data)
        self.assertIn("refresh", response.data)

    @override_settings(
        OTP_REST_AUTH={
            "VERIFICATION_REQUIRED": True,
            "VERIFICATION_METHOD": app_settings.AccountVerificationMethod.PHONE,
        }
    )
    def test_phone_login_with_phone_verification_type_verified(self):
        self.account.phone_verified = True
        self.account.save()

        data = {"phone": "+2348145640915", "password": "password"}
        response = self.client.post(self.url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn("access", response.data)
        self.assertIn("refresh", response.data)

    @override_settings(
        OTP_REST_AUTH={
            "VERIFICATION_REQUIRED": True,
            "VERIFICATION_METHOD": app_settings.AccountVerificationMethod.PHONE,
        }
    )
    def test_phone_login_with_phone_verification_type_unverified(self):
        data = {"phone": "+2348145640915", "password": "password"}
        response = self.client.post(self.url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertNotIn("access", response.data)
        self.assertNotIn("refresh", response.data)

    @override_settings(
        OTP_REST_AUTH={
            "VERIFICATION_REQUIRED": True,
            "VERIFICATION_METHOD": app_settings.AccountVerificationMethod.EMAIL,
        }
    )
    def test_email_login_with_account_verification_type_verified(self):
        self.account.email_verified = True
        self.account.save()

        data = {"email": self.user.email, "password": "password"}
        response = self.client.post(self.url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn("access", response.data)
        self.assertIn("refresh", response.data)

    @override_settings(
        OTP_REST_AUTH={
            "VERIFICATION_REQUIRED": True,
            "VERIFICATION_METHOD": app_settings.AccountVerificationMethod.EMAIL,
        }
    )
    def test_email_login_with_account_verification_type_unverified(self):
        data = {"email": self.user.email, "password": "password"}
        response = self.client.post(self.url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertNotIn("access", response.data)
        self.assertNotIn("refresh", response.data)

    @override_settings(
        OTP_REST_AUTH={
            "VERIFICATION_REQUIRED": True,
            "VERIFICATION_METHOD": app_settings.AccountVerificationMethod.PHONE,
        }
    )
    @skip("resolve later")
    def test_phone_login_without_phone_field(self):
        self.account.phone_verified = True
        self.account.save()

        data = {"email": self.user.email, "password": "password"}
        response = self.client.post(self.url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertNotIn("access", response.data)
        self.assertNotIn("refresh", response.data)

    @override_settings(
        OTP_REST_AUTH={
            "VERIFICATION_REQUIRED": True,
            "VERIFICATION_METHOD": app_settings.AccountVerificationMethod.EMAIL,
        }
    )
    @skip("resolve later")
    def test_email_login_without_email_field(self):
        self.account.email_verified = True
        self.account.save()

        data = {"phone": "+2348145640915", "password": "password"}
        response = self.client.post(self.url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertNotIn("access", response.data)
        self.assertNotIn("refresh", response.data)

    def test_post_inactive_user(self):
        self.user.is_active = False
        self.user.save()

        data = {"username": "testuser", "password": "password"}
        response = self.client.post(self.url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn("non_field_errors", response.data)
        self.assertEqual(
            response.data["non_field_errors"][0], "User account is disabled."
        )

    @override_settings(
        OTP_REST_AUTH={
            "VERIFICATION_REQUIRED": True,
            "VERIFICATION_METHOD": app_settings.AccountVerificationMethod.ACCOUNT,
        }
    )
    def test_post_unverified_account(self):
        self.account.is_verified = False
        self.account.save()

        data = {"username": "testuser", "password": "password"}
        response = self.client.post(self.url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn("non_field_errors", response.data)
        self.assertEqual(
            response.data["non_field_errors"][0], "Account is not verified."
        )

    @override_settings(
        OTP_REST_AUTH={
            "VERIFICATION_REQUIRED": False,
            "VERIFICATION_METHOD": app_settings.AccountVerificationMethod.ACCOUNT,
        }
    )
    def test_post_unverified_account_with_verification_not_required(self):
        self.account.is_verified = False
        self.account.save()

        data = {"username": "testuser", "password": "password"}
        response = self.client.post(self.url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn("access", response.data)
        self.assertIn("refresh", response.data)

    @override_settings(
        OTP_REST_AUTH={
            "VERIFICATION_REQUIRED": False,
            "VERIFICATION_METHOD": app_settings.AccountVerificationMethod.ACCOUNT,
            "AUTHENTICATION_METHODS": (
                app_settings.AuthenticationMethods.EMAIL,
                app_settings.AuthenticationMethods.USERNAME,
            ),
        }
    )
    def test_login_with_unavailable_authentication_type(self):
        data = {"phone": "+2348145640915", "password": "password"}
        response = self.client.post(self.url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertNotIn("access", response.data)
        self.assertNotIn("refresh", response.data)



from django.contrib.auth.backends import ModelBackend

from .app_settings import app_settings
from .utils import (
    get_auth_method_field,
    get_user_by_email,
    get_user_by_phone,
    get_user_by_username,
)


class AuthenticationBackend(ModelBackend):
    def authenticate(self, request, **credentials):
        user = None

        auth_methods = app_settings.AUTHENTICATION_METHODS
        for method in auth_methods:
            if not credentials.get(method):
                continue

            method_value = credentials.get(method)
            auth_method_field = get_auth_method_field(method)
            if auth_method_field == app_settings.USER_MODEL_USERNAME_FIELD:
                # username query is case sensitive if app_settings.PRESERVE_USERNAME_CASING
                # is set to True
                user = get_user_by_username(method_value)
            elif auth_method_field == app_settings.USER_MODEL_EMAIL_FIELD:
                user = get_user_by_email(method_value)
            elif auth_method_field == app_settings.USER_MODEL_PHONE_FIELD:
                user = get_user_by_phone(method_value)

            if not user or not user.check_password(credentials["password"]):
                return None

            return user



from django.apps import AppConfig


class OtpRestAuthConfig(AppConfig):
    name = "otp_rest_auth"
    default_auto_field = "django.db.models.BigAutoField"

    def ready(self):
        import otp_rest_auth.signals



from rest_framework import serializers
from django.contrib.auth.forms import SetPasswordForm
from django.utils.translation import gettext_lazy as _
from django.contrib.auth import get_user_model, authenticate
from django.core.exceptions import ObjectDoesNotExist
from phonenumber_field.serializerfields import PhoneNumberField
from django.core.exceptions import ValidationError as DjangoValidationError

from .app_settings import app_settings
from .models import Account, TOTP
from .otp_ops import verify_otp, validate_otp
from .utils import (
    get_user_by_phone,
    get_user_by_email,
    get_user_by_username,
    get_auth_method_field,
)


UserModel = get_user_model()
adapter = app_settings.ADAPTER()


class RegisterSerializer(serializers.Serializer):
    username = serializers.CharField(
        max_length=app_settings.USERNAME_MAX_LENGTH,
        min_length=app_settings.USERNAME_MIN_LENGTH,
        required=app_settings.USERNAME_REQUIRED,
    )
    phone = serializers.CharField()
    email = serializers.EmailField(required=app_settings.PHONE_REQUIRED)
    password1 = serializers.CharField(write_only=True)
    password2 = serializers.CharField(write_only=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not app_settings.SIGNUP_PASSWORD_ENTER_TWICE:
            self.fields.pop("password1")
            self.fields.pop("password2")
            self.fields["password"] = serializers.CharField(write_only=True)

        if (
            app_settings.AuthenticationMethods.PHONE
            not in app_settings.AUTHENTICATION_METHODS
        ):
            self.fields.pop("phone")

        if (
            app_settings.AuthenticationMethods.EMAIL
            not in app_settings.AUTHENTICATION_METHODS
        ):
            self.fields.pop("email")

        if (
            app_settings.AuthenticationMethods.USERNAME
            not in app_settings.AUTHENTICATION_METHODS
        ):
            self.fields.pop("username")

    def validate_username(self, username):
        username = adapter.clean_username(username)
        return username

    def validate_phone(self, phone):
        phone = adapter.clean_phone(phone)
        if app_settings.UNIQUE_PHONE:
            user = get_user_by_phone(phone)
            account = Account.objects.filter(user=user).first()

            if account and not account.is_verified:
                raise serializers.ValidationError(
                    _(
                        "Phone number is not verified. A verification code was sent to your phone. "
                        "Please verify your number or request a new code."
                    )
                )
            elif account and account.is_verified:
                raise serializers.ValidationError(
                    _("A user is already registered with this phone number."),
                )

        return phone

    def validate_email(self, email):
        email = adapter.clean_email(email)
        if app_settings.UNIQUE_EMAIL:
            user = get_user_by_email(email)
            account = Account.objects.filter(user=user).first()

            if account and not account.is_verified:
                raise serializers.ValidationError(
                    _(
                        "Email address is not verified. A verification code was sent to your email. "
                        "Please verify your email or request a new code."
                    )
                )
            elif account and account.is_verified:
                raise serializers.ValidationError(
                    _("A user is already registered with this e-mail address."),
                )

        return email

    def validate_password1(self, password):
        if app_settings.SIGNUP_PASSWORD_VERIFICATION:
            return adapter.clean_password(password)

    def validate(self, data):
        if app_settings.SIGNUP_PASSWORD_ENTER_TWICE:
            if data["password1"] != data["password2"]:
                raise serializers.ValidationError(
                    _("The two password fields didn't match.")
                )

        if (
            app_settings.VERIFICATION_METHOD
            == app_settings.AccountVerificationMethod.ACCOUNT
        ):
            user_from_email = get_user_by_email(data.get("email"))
            user_from_phone = get_user_by_phone(data.get("phone"))
            if user_from_email or user_from_phone:
                raise serializers.ValidationError(
                    _(
                        "An account already exists with the provided email or phone number. "
                        "Please verify your account or use a different email or phone number."
                    )
                )

        return data

    def get_cleaned_data(self):
        return {
            "username": self.validated_data.get("username", ""),
            "password1": self.validated_data.get("password1", ""),
            "email": self.validated_data.get("email", ""),
            "phone": self.validated_data.get("phone", ""),
        }

    def save(self, request):
        user = adapter.new_user(request)
        self.cleaned_data = self.get_cleaned_data()
        user = adapter.save_user(request, user, self, commit=False)
        if "password1" in self.cleaned_data:
            try:
                adapter.clean_password(self.cleaned_data["password1"], user=user)
            except DjangoValidationError as exc:
                raise serializers.ValidationError(
                    detail=serializers.as_serializer_error(exc)
                )

        if app_settings.VERIFICATION_REQUIRED:
            user.is_active = False

        user.save()
        return user


class OTPSerializer(serializers.Serializer):
    otp = serializers.IntegerField()


class ResendOTPSerializer(serializers.Serializer):
    phone = serializers.CharField(required=False)
    email = serializers.EmailField(required=False)
    purpose = serializers.ChoiceField(choices=TOTP.PURPOSE_CHOICES, required=True)

    def validate_phone(self, phone):
        return adapter.clean_phone(phone)

    def validate(self, data):
        if data["purpose"] == TOTP.PURPOSE_ACCOUNT_VERIFICATION:
            if "phone" not in data and "email" not in data:
                raise serializers.ValidationError(
                    _("Either 'phone' or 'email' field is required.")
                )

        if data["purpose"] == TOTP.PURPOSE_EMAIL_VERIFICATION:
            if "email" not in data:
                raise serializers.ValidationError(_('"email" field is required.'))

        if data["purpose"] == TOTP.PURPOSE_PHONE_VERIFICATION:
            if "phone" not in data:
                raise serializers.ValidationError(_('"phone" field is required.'))

        if data["purpose"] == TOTP.PURPOSE_PASSWORD_RESET:
            if (
                "phone" in app_settings.PASSWORD_RESET_OTP_RECIPIENTS
                and "email" in app_settings.PASSWORD_RESET_OTP_RECIPIENTS
            ):
                if "phone" not in data and "email" not in data:
                    raise serializers.ValidationError(
                        _("Either 'phone' or 'email' field is required.")
                    )
            elif "phone" in app_settings.PASSWORD_RESET_OTP_RECIPIENTS:
                if "phone" not in data:
                    raise serializers.ValidationError(_("'phone' field is required."))
            elif "email" in app_settings.PASSWORD_RESET_OTP_RECIPIENTS:
                if "email" not in data:
                    raise serializers.ValidationError(_("'email' field is required."))

        return super().validate(data)


class UserDetailsSerializer(serializers.ModelSerializer):
    """
    User model w/o password
    """

    @staticmethod
    def validate_username(username):
        username = adapter.clean_username(username)
        return username

    class Meta:
        extra_fields = []
        # see https://github.com/iMerica/dj-rest-auth/issues/181
        # UserModel.XYZ causing attribute error while importing other
        # classes from `serializers.py`. So, we need to check whether the auth model has
        # the attribute or not
        if hasattr(UserModel, "USERNAME_FIELD"):
            extra_fields.append(UserModel.USERNAME_FIELD)
        if hasattr(UserModel, "EMAIL_FIELD"):
            extra_fields.append(UserModel.EMAIL_FIELD)
        if hasattr(UserModel, app_settings.USER_MODEL_PHONE_FIELD):
            extra_fields.append(app_settings.USER_MODEL_PHONE_FIELD)
        if hasattr(UserModel, "first_name"):
            extra_fields.append("first_name")
        if hasattr(UserModel, "last_name"):
            extra_fields.append("last_name")
        model = UserModel
        fields = ("pk", *extra_fields)
        read_only_fields = ("email", "phone")


class JWTSerializer(serializers.Serializer):
    """
    Serializer for JWT authentication.
    """

    access = serializers.CharField()
    refresh = serializers.CharField()
    user = serializers.SerializerMethodField()

    def get_user(self, obj):
        """
        Required to allow using custom USER_DETAILS_SERIALIZER in
        JWTSerializer. Defining it here to avoid circular imports
        """
        JWTUserDetailsSerializer = app_settings.USER_DETAILS_SERIALIZER

        user_data = JWTUserDetailsSerializer(obj["user"], context=self.context).data
        return user_data


class JWTSerializerWithExpiration(JWTSerializer):
    """
    Serializer for JWT authentication with expiration times.
    """

    access_expiration = serializers.DateTimeField()
    refresh_expiration = serializers.DateTimeField()


class LoginSerializer(serializers.Serializer):
    phone = serializers.CharField(required=False, allow_blank=True)
    email = serializers.EmailField(required=False, allow_blank=True)
    username = serializers.CharField(required=False, allow_blank=True)
    password = serializers.CharField(style={"input_type": "password"})

    def __init__(self, instance=None, data=..., **kwargs):
        super().__init__(instance, data, **kwargs)

        if (
            app_settings.AuthenticationMethods.USERNAME
            not in app_settings.AUTHENTICATION_METHODS
        ):
            self.fields.pop("username")
        if (
            app_settings.AuthenticationMethods.PHONE
            not in app_settings.AUTHENTICATION_METHODS
        ):
            self.fields.pop("phone")
        if (
            app_settings.AuthenticationMethods.EMAIL
            not in app_settings.AUTHENTICATION_METHODS
        ):
            self.fields.pop("email")

    def authenticate(self, **kwargs):
        return authenticate(self.context["request"], **kwargs)

    def validate_phone(self, phone):
        return adapter.clean_phone(phone)

    def get_user(self, data):
        auth_methods = app_settings.AUTHENTICATION_METHODS
        credentials = {}

        for method in auth_methods:
            if data.get(method) and data.get("password"):
                credentials[method] = data.get(method)
                credentials["password"] = data.get("password")
                break

        if not credentials:
            if len(auth_methods) == 1:
                msg = _(f'Must include "{auth_methods[0]}" and "password".')
            else:
                auth_methods_str = (
                    '", "'.join(auth_methods[:-1]) + f'", or "{auth_methods[-1]}"'
                )
                auth_methods_str = f'"{auth_methods_str}'
                msg = _(f'Must include either {auth_methods_str} and "password".')
            raise serializers.ValidationError(msg)

        user = self.authenticate(**credentials)
        return user

    @staticmethod
    def validate_auth_user_status(user):
        if not user.is_active:
            msg = _("User account is disabled.")
            raise serializers.ValidationError(msg)

    @staticmethod
    def validate_verification_type_status(user):
        user_account = Account.objects.filter(user=user).first()
        if not user_account:
            raise ObjectDoesNotExist(
                "The related Account instance for this user does not exist."
            )

        if app_settings.VERIFICATION_REQUIRED:
            verification_type = app_settings.VERIFICATION_METHOD
            if (
                verification_type == app_settings.AccountVerificationMethod.ACCOUNT
                and not user_account.is_verified
            ):
                raise serializers.ValidationError(_("Account is not verified."))
            if (
                verification_type == app_settings.AccountVerificationMethod.EMAIL
                and not user_account.email_verified
            ):
                raise serializers.ValidationError(_("E-mail is not verified."))
            if (
                verification_type == app_settings.AccountVerificationMethod.PHONE
                and not user_account.phone_verified
            ):
                raise serializers.ValidationError(_("Phone number is not verified."))

    def validate(self, attrs):
        user = self.get_user(attrs)
        if not user:
            msg = _("Unable to log in with provided credentials.")
            raise serializers.ValidationError(msg)

        self.validate_verification_type_status(user)
        self.validate_auth_user_status(user)

        attrs["user"] = user
        return attrs


class LogoutSerializer(serializers.Serializer):
    refresh = serializers.CharField()


class PasswordResetSerializer(serializers.Serializer):
    phone = serializers.CharField(required=False, allow_blank=True)
    email = serializers.EmailField(required=False, allow_blank=True)
    username = serializers.CharField(required=False, allow_blank=True)

    def get_user(self, data):
        user = None
        requried_auth_method_provided = False

        auth_methods = app_settings.AUTHENTICATION_METHODS
        for method in auth_methods:
            if data.get(method):
                method_value = data.get(method)
                if method_value:
                    requried_auth_method_provided = True
                else:
                    continue

                method_field = get_auth_method_field(method)

                if method_field == app_settings.USER_MODEL_USERNAME_FIELD:
                    user = get_user_by_username(method_value)
                elif method_field == app_settings.USER_MODEL_EMAIL_FIELD:
                    user = get_user_by_email(method_value)
                elif method_field == app_settings.USER_MODEL_PHONE_FIELD:
                    user = get_user_by_phone(method_value)

                if user:
                    break

        if not requried_auth_method_provided:
            if len(auth_methods) == 1:
                msg = _(f'Must include "{auth_methods[0]}".')
            else:
                auth_methods_str = (
                    '", "'.join(auth_methods[:-1]) + f'", or "{auth_methods[-1]}"'
                )
                auth_methods_str = f'"{auth_methods_str}'
                msg = _(f"Must include either {auth_methods_str}.")
            raise serializers.ValidationError(msg)

        return user

    def validate_phone(self, phone):
        return adapter.clean_phone(phone)

    def validate(self, attrs):
        user = self.get_user(attrs)
        attrs["user"] = user

        return attrs


class PasswordResetConfirmSerializer(serializers.Serializer):
    otp = serializers.IntegerField()
    new_password1 = serializers.CharField(max_length=128)
    new_password2 = serializers.CharField(max_length=128)

    totp = None
    set_password_form = None

    def validate_otp(self, otp):
        try:
            is_valid, self.totp = validate_otp(otp, TOTP.PURPOSE_PASSWORD_RESET)
            if not is_valid:
                raise serializers.ValidationError(_("Invalid OTP."))

        except TOTP.DoesNotExist:
            raise serializers.ValidationError(_("Invalid OTP."))

    def validate(self, data):
        self.set_password_form = SetPasswordForm(
            user=self.totp.user,
            data=data,
        )
        if not self.set_password_form.is_valid():
            raise serializers.ValidationError(self.set_password_form.errors)

        return data

    def save(self):
        verify_otp(self.totp.otp, TOTP.PURPOSE_PASSWORD_RESET)
        return self.set_password_form.save()


class PasswordChangeSerializer(serializers.Serializer):
    refresh = serializers.CharField()
    old_password = serializers.CharField(max_length=128)
    new_password1 = serializers.CharField(max_length=128)
    new_password2 = serializers.CharField(max_length=128)

    set_password_form_class = SetPasswordForm

    set_password_form = None

    def __init__(self, *args, **kwargs):
        self.old_password_field_enabled = app_settings.OLD_PASSWORD_FIELD_ENABLED
        self.logout_on_password_change = app_settings.LOGOUT_ON_PASSWORD_CHANGE
        super().__init__(*args, **kwargs)

        if not self.old_password_field_enabled:
            self.fields.pop("old_password")

        if not self.logout_on_password_change:
            self.fields.pop("refresh")

        self.request = self.context.get("request")
        self.user = getattr(self.request, "user", None)

    def validate_old_password(self, value):
        invalid_password_conditions = (
            self.old_password_field_enabled,
            self.user,
            not self.user.check_password(value),
        )

        if all(invalid_password_conditions):
            err_msg = _(
                "Your old password was entered incorrectly. Please enter it again."
            )
            raise serializers.ValidationError(err_msg)
        return value

    def validate(self, attrs):
        self.set_password_form = self.set_password_form_class(
            user=self.user,
            data=attrs,
        )

        if not self.set_password_form.is_valid():
            raise serializers.ValidationError(self.set_password_form.errors)
        return attrs

    def save(self):
        self.set_password_form.save()


class ChangeEmailSerializer(serializers.Serializer):
    new_email = serializers.EmailField()

    def validate_new_email(self, new_email):
        email_field = app_settings.USER_MODEL_EMAIL_FIELD
        if app_settings.UNIQUE_EMAIL:
            if UserModel.objects.filter(**{email_field: new_email}).exists():
                raise serializers.ValidationError("Email address already exists.")
        return new_email


class ChangePhoneSerializer(serializers.Serializer):
    new_phone = PhoneNumberField()

    def validate_new_phone(self, new_phone):
        phone_field = app_settings.USER_MODEL_PHONE_FIELD
        if app_settings.UNIQUE_PHONE:
            if UserModel.objects.filter(**{phone_field: new_phone}).exists():
                raise serializers.ValidationError("Phone number already exists.")
        return new_phone


class ChangeEmailConfirmSerializer(serializers.Serializer):
    otp = serializers.IntegerField()


class ChangePhoneConfirmSerializer(serializers.Serializer):
    otp = serializers.IntegerField()



from rest_framework import status
from rest_framework.test import APIClient
from unittest.mock import patch, MagicMock
from django.contrib.auth import get_user_model
from django.test.utils import override_settings
from django.test import TestCase, RequestFactory
from rest_framework_simplejwt.tokens import AccessToken, RefreshToken

from otp_rest_auth.models import Account, TOTP
from otp_rest_auth.serializers import JWTSerializer
from otp_rest_auth.app_settings import app_settings
from otp_rest_auth.views import get_login_response_data, verify

User = get_user_model()


class GetLoginResponseDataTests(TestCase):
    def setUp(self):
        self.factory = RequestFactory()
        self.user = User.objects.create_user(username="testuser", password="12345")
        self.context = {"request": self.factory.get("/")}

    @patch("otp_rest_auth.jwt_auth.get_tokens_for_user")
    @override_settings(OTP_REST_AUTH={"JWT_AUTH_RETURN_EXPIRATION": True})
    def test_get_login_response_data_with_expiration(self, mock_get_tokens):
        mock_get_tokens.return_value = ("access_token", "refresh_token")

        # with patch.object(app_settings, "JWT_AUTH_RETURN_EXPIRATION", True):
        data = get_login_response_data(self.user, self.context)

        self.assertIn("access", data)
        self.assertIn("refresh", data)
        self.assertIn("access_expiration", data)
        self.assertIn("refresh_expiration", data)

    @patch("otp_rest_auth.jwt_auth.get_tokens_for_user")
    @override_settings(OTP_REST_AUTH={"JWT_AUTH_RETURN_EXPIRATION": False})
    def test_get_login_response_data_without_expiration(self, mock_get_tokens):
        mock_get_tokens.return_value = ("access_token", "refresh_token")

        data = get_login_response_data(self.user, self.context)

        self.assertIn("access", data)
        self.assertIn("refresh", data)
        self.assertNotIn("access_expiration", data)
        self.assertNotIn("refresh_expiration", data)

    @patch("otp_rest_auth.jwt_auth.get_tokens_for_user")
    @override_settings(OTP_REST_AUTH={"JWT_AUTH_HTTPONLY": True})
    def test_get_login_response_data_with_httponly(self, mock_get_tokens):
        mock_get_tokens.return_value = ("access_token", "refresh_token")

        data = get_login_response_data(self.user, self.context)

        self.assertIn("access", data)
        self.assertIn("refresh", data)
        self.assertEqual(data["refresh"], "")

    @patch("otp_rest_auth.jwt_auth.get_tokens_for_user")
    @override_settings(OTP_REST_AUTH={"JWT_AUTH_HTTPONLY": False})
    def test_get_login_response_data_without_httponly(self, mock_get_tokens):
        mock_get_tokens.return_value = ("access_token", "refresh_token")

        data = get_login_response_data(self.user, self.context)

        self.assertIn("access", data)
        self.assertIn("refresh", data)
        self.assertNotEqual(data["refresh"], "")

    @patch("otp_rest_auth.jwt_auth.get_tokens_for_user")
    @override_settings(
        OTP_REST_AUTH={
            "JWT_AUTH_HTTPONLY": False,
            "JWT_AUTH_RETURN_EXPIRATION": True,
            "JWT_SERIALIZER": JWTSerializer,
        }
    )
    def test_get_login_response_data_serializer(self, mock_get_tokens):
        mock_get_tokens.return_value = ("access_token", "refresh_token")

        data = get_login_response_data(self.user, self.context)

        self.assertIsInstance(data, dict)
        self.assertIn("user", data)
        self.assertIn("access", data)
        self.assertIn("refresh", data)


class VerifyTests(TestCase):
    def setUp(self):
        self.client = APIClient()
        self.factory = RequestFactory()
        self.user = User.objects.create_user(username="testuser", password="12345")
        self.account = Account.objects.get(user=self.user)

    def test_verify_success(self):
        totp = TOTP.objects.create(
            user=self.user, purpose=TOTP.PURPOSE_ACCOUNT_VERIFICATION
        )
        serializer = MagicMock()
        serializer.validated_data = {"otp": totp.otp}

        response = verify(
            serializer, self.factory.get("/"), TOTP.PURPOSE_ACCOUNT_VERIFICATION
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.account.refresh_from_db()
        self.assertTrue(self.account.is_verified)
        self.assertTrue(self.user.is_active)

    def test_verify_invalid_otp(self):
        serializer = MagicMock()
        serializer.validated_data = {"otp": None}

        response = verify(
            serializer, self.factory.get("/"), TOTP.PURPOSE_ACCOUNT_VERIFICATION
        )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data["detail"], "Invalid OTP.")

    @override_settings(OTP_REST_AUTH={"LOGIN_UPON_VERIFICATION": True})
    def test_verify_login_upon_verification(self):
        totp = TOTP.objects.create(
            user=self.user, purpose=TOTP.PURPOSE_ACCOUNT_VERIFICATION
        )
        serializer = MagicMock()
        serializer.validated_data = {"otp": totp.otp}

        response = verify(
            serializer,
            self.factory.get("/"),
            TOTP.PURPOSE_ACCOUNT_VERIFICATION,
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn("access", response.data)
        self.assertIn("refresh", response.data)

    @override_settings(OTP_REST_AUTH={"LOGIN_UPON_VERIFICATION": False})
    def test_verify_no_login_upon_verification(self):
        totp = TOTP.objects.create(
            user=self.user, purpose=TOTP.PURPOSE_ACCOUNT_VERIFICATION
        )
        serializer = MagicMock()
        serializer.validated_data = {"otp": totp.otp}

        response = verify(
            serializer,
            self.factory.get("/"),
            TOTP.PURPOSE_ACCOUNT_VERIFICATION,
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNone(response.data)



from django.test import TestCase, RequestFactory
from rest_framework.test import APIClient
from unittest.mock import patch
from rest_framework import status
from rest_framework.response import Response
from django.contrib.auth import get_user_model
from django.test.utils import override_settings

from otp_rest_auth.app_settings import app_settings
from otp_rest_auth.models import TOTP
from otp_rest_auth.serializers import OTPSerializer
from otp_rest_auth.views import VerifyAccountView, VerifyEmailView, VerifyPhoneView

User = get_user_model()


class VerifyAccountViewTests(TestCase):
    def setUp(self):
        self.client = APIClient()
        self.factory = RequestFactory()
        self.view = VerifyAccountView.as_view()
        self.user = User.objects.create_user(username="testuser", password="12345")
        self.totp = TOTP.objects.create(
            user=self.user, purpose=TOTP.PURPOSE_ACCOUNT_VERIFICATION
        )

    def test_get_not_allowed(self):
        request = self.factory.get("/verify-account/")
        response = self.view(request)
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    @patch("otp_rest_auth.views.verify")
    @override_settings(OTP_REST_AUTH={"OTP_SERIALIZER": OTPSerializer})
    def test_post_valid_data(self, mock_verify):
        data = {"otp": self.totp.otp}
        request = self.factory.post("/verify-account/", data, format="json")

        mock_response = Response(
            {"detail": "Verification successful."}, status=status.HTTP_200_OK
        )
        mock_verify.return_value = mock_response

        response = self.view(request)
        mock_verify.assert_called_once
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {"detail": "Verification successful."})

    @patch("otp_rest_auth.views.verify")
    @override_settings(OTP_REST_AUTH={"OTP_SERIALIZER": OTPSerializer})
    def test_post_invalid_data(self, mock_verify):
        data = {"otp": "wrong_otp"}
        request = self.factory.post("/verify-account/", data, format="json")

        serializer = app_settings.OTP_SERIALIZER(data=data)
        self.assertFalse(serializer.is_valid())

        response = self.view(request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        mock_verify.assert_not_called()


class VerifyEmailViewTests(TestCase):
    def setUp(self):
        self.client = APIClient()
        self.factory = RequestFactory()
        self.view = VerifyAccountView.as_view()
        self.user = User.objects.create_user(username="testuser", password="12345")
        self.totp = TOTP.objects.create(
            user=self.user, purpose=TOTP.PURPOSE_PHONE_VERIFICATION
        )

    def test_get_not_allowed(self):
        request = self.factory.get("/verify-account/")
        response = self.view(request)
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    @patch("otp_rest_auth.views.verify")
    @override_settings(OTP_REST_AUTH={"OTP_SERIALIZER": OTPSerializer})
    def test_post_valid_data(self, mock_verify):
        data = {"otp": self.totp.otp}
        request = self.factory.post("/verify-account/", data, format="json")

        mock_response = Response(
            {"detail": "Verification successful."}, status=status.HTTP_200_OK
        )
        mock_verify.return_value = mock_response

        response = self.view(request)
        mock_verify.assert_called_once
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {"detail": "Verification successful."})

    @patch("otp_rest_auth.views.verify")
    @override_settings(OTP_REST_AUTH={"OTP_SERIALIZER": OTPSerializer})
    def test_post_invalid_data(self, mock_verify):
        data = {"otp": "wrong_otp"}
        request = self.factory.post("/verify-account/", data, format="json")

        serializer = app_settings.OTP_SERIALIZER(data=data)
        self.assertFalse(serializer.is_valid())

        response = self.view(request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        mock_verify.assert_not_called()


class VerifyPhoneViewTests(TestCase):
    def setUp(self):
        self.client = APIClient()
        self.factory = RequestFactory()
        self.view = VerifyAccountView.as_view()
        self.user = User.objects.create_user(username="testuser", password="12345")
        self.totp = TOTP.objects.create(
            user=self.user, purpose=TOTP.PURPOSE_PHONE_VERIFICATION
        )

    def test_get_not_allowed(self):
        request = self.factory.get("/verify-account/")
        response = self.view(request)
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    @patch("otp_rest_auth.views.verify")
    @override_settings(OTP_REST_AUTH={"OTP_SERIALIZER": OTPSerializer})
    def test_post_valid_data(self, mock_verify):
        data = {"otp": self.totp.otp}
        request = self.factory.post("/verify-account/", data, format="json")

        mock_response = Response(
            {"detail": "Verification successful."}, status=status.HTTP_200_OK
        )
        mock_verify.return_value = mock_response

        response = self.view(request)
        mock_verify.assert_called_once
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {"detail": "Verification successful."})

    @patch("otp_rest_auth.views.verify")
    @override_settings(OTP_REST_AUTH={"OTP_SERIALIZER": OTPSerializer})
    def test_post_invalid_data(self, mock_verify):
        data = {"otp": "wrong_otp"}
        request = self.factory.post("/verify-account/", data, format="json")

        serializer = app_settings.OTP_SERIALIZER(data=data)
        self.assertFalse(serializer.is_valid())

        response = self.view(request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        mock_verify.assert_not_called()



from datetime import timedelta
from django.test import TestCase
from django.contrib.auth import get_user_model
from django.test.utils import override_settings
from django.utils import timezone
from django.core.exceptions import ValidationError

from otp_rest_auth.models import Account, TOTP
from otp_rest_auth.app_settings import app_settings


User = get_user_model()


class AccountModelTests(TestCase):
    def setUp(self):
        self.user = User.objects.create_user(username="testuser", password="password")
        self.account = Account.objects.get(user=self.user)

    def test_initial_account_creation(self):
        account = Account.objects.get(user=self.user)
        self.assertFalse(account.is_verified)
        self.assertFalse(account.phone_verified)
        self.assertFalse(account.email_verified)
        self.assertIsNone(account.verified_at)
        self.assertIsNone(account.phone_verified_at)
        self.assertIsNone(account.email_verified_at)

    def test_phone_verified_sets_is_verified(self):
        self.account.phone_verified = True
        self.account.save()
        self.account.refresh_from_db()

        self.assertTrue(self.account.is_verified)
        self.assertIsNotNone(self.account.verified_at)
        self.assertIsNotNone(self.account.phone_verified_at)
        self.assertIsNone(self.account.email_verified_at)

    def test_email_verified_sets_is_verified(self):
        self.account.email_verified = True
        self.account.save()
        self.account.refresh_from_db()

        self.assertTrue(self.account.is_verified)
        self.assertIsNotNone(self.account.verified_at)
        self.assertIsNotNone(self.account.email_verified_at)
        self.assertIsNone(self.account.phone_verified_at)

    def test_phone_and_email_verified(self):
        self.account.phone_verified = True
        self.account.email_verified = True
        self.account.save()
        self.account.refresh_from_db()

        self.assertTrue(self.account.is_verified)
        self.assertIsNotNone(self.account.verified_at)
        self.assertIsNotNone(self.account.phone_verified_at)
        self.assertIsNotNone(self.account.email_verified_at)

    def test_is_verified_set_does_not_override(self):
        self.account.is_verified = True
        self.account.save()
        initial_verified_at = self.account.verified_at

        self.account.save()
        self.account.refresh_from_db()

        self.assertTrue(self.account.is_verified)
        self.assertEqual(self.account.verified_at, initial_verified_at)

    def test_unverified_phone_and_email_does_not_set_is_verified(self):
        self.account.phone_verified = False
        self.account.email_verified = False
        self.account.save()
        self.account.refresh_from_db()

        self.assertFalse(self.account.is_verified)
        self.assertIsNone(self.account.verified_at)
        self.assertIsNone(self.account.phone_verified_at)
        self.assertIsNone(self.account.email_verified_at)


class TOTPModelTests(TestCase):
    @override_settings(
        OTP_REST_AUTH={
            "OTP_LENGTH": 4,
            "OTP_EXPIRY_TIME": 60,
            "PASSWORD_RESET_OTP_EXPIRY_TIME ": 90,
        }
    )
    def setUp(self):
        self.user = User.objects.create_user(username="testuser", password="password")
        self.totp = TOTP.objects.create(
            user=self.user,
            purpose=TOTP.PURPOSE_EMAIL_VERIFICATION,
        )

    def test_initial_totp_creation(self):
        totp = TOTP.objects.create(
            user=self.user,
            purpose=TOTP.PURPOSE_PASSWORD_RESET,
        )
        self.assertEqual(totp.user, self.user)
        self.assertTrue(totp.is_valid)
        self.assertIsNone(totp.invalidated_at)
        self.assertEqual(totp.purpose, TOTP.PURPOSE_PASSWORD_RESET)
        self.assertIsNotNone(totp.expiration_time)
        self.assertIsNotNone(totp.created_at)

    @override_settings(OTP_REST_AUTH={"OTP_LENGTH": 9})
    def test_otp_generation(self):
        totp = TOTP.objects.create(
            user=self.user,
            purpose=TOTP.PURPOSE_EMAIL_VERIFICATION,
        )
        self.assertNotEqual(self.totp.otp, totp.otp)
        self.assertEqual(len(str(totp.otp)), app_settings.OTP_LENGTH)

    def test_is_valid_reset_to_true(self):
        totp = TOTP.objects.create(
            user=self.user,
            purpose=TOTP.PURPOSE_PHONE_VERIFICATION,
            is_valid=False,
        )
        totp.is_valid = True
        with self.assertRaises(ValidationError):
            totp.clean()

    def test_otp_update(self):
        totp = TOTP.objects.create(
            user=self.user,
            purpose=TOTP.PURPOSE_ACCOUNT_VERIFICATION,
        )
        totp.otp = 123456
        with self.assertRaises(ValidationError):
            totp.clean()

    def test_invalidated_at_set(self):
        totp = TOTP.objects.create(
            user=self.user,
            purpose=TOTP.PURPOSE_PHONE_VERIFICATION,
            is_valid=False,
        )
        self.assertIsNotNone(totp.invalidated_at)

    def test_is_expired(self):
        totp_password_reset = TOTP.objects.create(
            user=self.user,
            is_valid=True,
            purpose=TOTP.PURPOSE_PASSWORD_RESET,
            expiration_time=timezone.now() + timedelta(seconds=3600),  # 1 hour from now
        )
        totp_email_verification = TOTP.objects.create(
            user=self.user,
            is_valid=True,
            purpose=TOTP.PURPOSE_EMAIL_VERIFICATION,
            expiration_time=timezone.now() + timedelta(seconds=3600),  # 1 hour from now
        )

        self.assertFalse(totp_password_reset.is_expired)
        self.assertFalse(totp_email_verification.is_expired)

        expired_totp = TOTP.objects.create(
            user=self.user,
            is_valid=True,
            purpose=TOTP.PURPOSE_PHONE_VERIFICATION,
            expiration_time=timezone.now() - timedelta(seconds=3600),  # 1 hour ago
        )
        self.assertTrue(expired_totp.is_expired)

    @override_settings(
        OTP_REST_AUTH={"OTP_EXPIRY_TIME": 60, "PASSWORD_RESET_OTP_EXPIRY_TIME": 120}
    )
    def test_expiration_time_calculation(self):
        # Test Password Reset OTP expiration time calculation
        password_reset_otp = TOTP.objects.create(
            user=self.user,
            is_valid=True,
            purpose=TOTP.PURPOSE_PASSWORD_RESET,
        )
        prt = timezone.now() + timedelta(
            seconds=app_settings.PASSWORD_RESET_OTP_EXPIRY_TIME
        )
        self.assertEqual(password_reset_otp.expiration_time.second, prt.second)

        # Test Email Verification OTP expiration time calculation
        email_verification_otp = TOTP.objects.create(
            user=self.user,
            is_valid=True,
            purpose=TOTP.PURPOSE_EMAIL_VERIFICATION,
        )
        evt = timezone.now() + timedelta(seconds=app_settings.OTP_EXPIRY_TIME)
        self.assertEqual(email_verification_otp.expiration_time.second, evt.second)

        self.assertNotEqual(prt.minute, evt.minute)



from django.urls import path, include

from .views import (
    RegisterView,
    ResendOTPView,
    ResetPasswordView,
    LoginView,
    LogoutView,
    UserDetailsView,
    VerifyAccountView,
    VerifyEmailView,
    VerifyPhoneView,
    PasswordChangeView,
    PasswordResetConfirmView,
    InvokeChangeEmailOTPView,
    InvokeChangePhoneOTPView,
    ChangeEmailConfrimationView,
    ChangePhoneConfirmationView,
)
from rest_framework_simplejwt.views import TokenVerifyView
from otp_rest_auth.jwt_auth import get_refresh_view

urlpatterns = [
    path("login/", LoginView.as_view(), name="otp_rest_login"),
    path("logout/", LogoutView.as_view(), name="otp_rest_logout"),
    path("register/", RegisterView.as_view(), name="otp_rest_register"),
    path("user/", UserDetailsView.as_view(), name="otp_rest_user_details"),
    path("resend_otp/", ResendOTPView.as_view(), name="otp_rest_resend_otp"),
    path(
        "phone/change/",
        InvokeChangePhoneOTPView.as_view(),
        name="otp_rest_change_phone",
    ),
    path(
        "email/change/",
        InvokeChangeEmailOTPView.as_view(),
        name="otp_rest_change_email",
    ),
    path(
        "phone/change/confirm/",
        ChangePhoneConfirmationView.as_view(),
        name="otp_rest_change_phone_set",
    ),
    path(
        "email/change/confirm/",
        ChangeEmailConfrimationView.as_view(),
        name="otp_rest_change_email_set",
    ),
    path("verify/phone/", VerifyPhoneView.as_view(), name="otp_rest_verify_phone"),
    path("verify/email/", VerifyEmailView.as_view(), name="otp_rest_verify_email"),
    path(
        "verify/account/", VerifyAccountView.as_view(), name="otp_rest_verify_account"
    ),
    path(
        "password/reset/", ResetPasswordView.as_view(), name="otp_rest_password_reset"
    ),
    path(
        "password/reset/confirm/",
        PasswordResetConfirmView.as_view(),
        name="otp_rest_password_reset_confirm",
    ),
    path(
        "password/change/",
        PasswordChangeView.as_view(),
        name="otp_rest_password_change",
    ),
    path("token/verify/", TokenVerifyView.as_view(), name="token_verify"),
    path("token/refresh/", get_refresh_view().as_view(), name="token_refresh"),
]



from django.test import TestCase
from django.contrib.auth import get_user_model
from django.test.utils import override_settings

from otp_rest_auth.app_settings import app_settings
from otp_rest_auth.auth_backends import AuthenticationBackend


class AuthenticationBackendTests(TestCase):
    def setUp(self):
        user = get_user_model().objects.create(
            is_active=True,
            email="john@example.com",
            username="john",
            phone="+2348076048994",
        )
        user.set_password(user.username)
        user.save()
        self.user = user

    @override_settings(
        OTP_REST_AUTH={
            "AUTHENTICATION_METHODS": [app_settings.AuthenticationMethods.USERNAME]
        }
    )
    def test_auth_by_username(self):
        user = self.user
        backend = AuthenticationBackend()
        self.assertEqual(
            backend.authenticate(
                request=None, username=user.username, password=user.username
            ).pk,
            user.pk,
        )
        self.assertEqual(
            backend.authenticate(
                request=None, username=user.email, password=user.username
            ),
            None,
        )

    @override_settings(
        OTP_REST_AUTH={
            "AUTHENTICATION_METHODS": [app_settings.AuthenticationMethods.EMAIL]
        }
    )
    def test_auth_by_email(self):
        user = self.user
        backend = AuthenticationBackend()
        self.assertEqual(
            backend.authenticate(
                request=None, email=user.email, password=user.username
            ).pk,
            user.pk,
        )
        self.assertEqual(
            backend.authenticate(
                request=None, email=user.username, password=user.username
            ),
            None,
        )

    @override_settings(
        OTP_REST_AUTH={
            "AUTHENTICATION_METHODS": [app_settings.AuthenticationMethods.PHONE]
        }
    )
    def test_auth_by_phone(self):
        user = self.user
        backend = AuthenticationBackend()
        self.assertEqual(
            backend.authenticate(
                request=None, phone=user.phone, password=user.username
            ).pk,
            user.pk,
        )
        self.assertEqual(
            backend.authenticate(
                request=None, phone=user.username, password=user.username
            ),
            None,
        )

    @override_settings(
        OTP_REST_AUTH={
            "AUTHENTICATION_METHODS": [
                app_settings.AuthenticationMethods.PHONE,
                app_settings.AuthenticationMethods.EMAIL,
                app_settings.AuthenticationMethods.USERNAME,
            ]
        }
    )
    def test_auth_by_all_methods(self):
        user = self.user
        backend = AuthenticationBackend()
        self.assertEqual(
            backend.authenticate(
                request=None, username=user.username, password=user.username
            ).pk,
            user.pk,
        )
        self.assertEqual(
            backend.authenticate(
                request=None, email=user.email, password=user.username
            ).pk,
            user.pk,
        )
        self.assertEqual(
            backend.authenticate(
                request=None, phone=user.phone, password=user.username
            ).pk,
            user.pk,
        )



from django.conf import settings
from django.utils import timezone
from django.contrib.auth import get_user_model
from django.utils.decorators import method_decorator
from django.utils.translation import gettext_lazy as _
from django.views.decorators.debug import sensitive_post_parameters
from rest_framework import status, views
from rest_framework.response import Response
from rest_framework.exceptions import MethodNotAllowed
from rest_framework_simplejwt.tokens import RefreshToken
from rest_framework_simplejwt.exceptions import TokenError
from rest_framework.permissions import AllowAny, IsAuthenticated
from rest_framework.generics import CreateAPIView, GenericAPIView, RetrieveUpdateAPIView

from .app_settings import app_settings
from . import signals
from .utils import get_throttle_scope
from .models import Account, TOTP, TOTPMetadata
from .otp_ops import send_verification_otp, verify_otp

from .jwt_auth import get_tokens_for_user, set_jwt_cookies, unset_jwt_cookies


UserModel = get_user_model()
adapter = app_settings.ADAPTER()
sensitive_post_parameters_m = method_decorator(
    sensitive_post_parameters("password1", "password2"),
)


def get_login_response_data(user, context):
    from rest_framework_simplejwt.settings import (
        api_settings as jwt_settings,
    )

    serializer_class = app_settings.JWT_SERIALIZER
    if app_settings.JWT_AUTH_RETURN_EXPIRATION:
        serializer_class = app_settings.JWT_SERIALIZER_WITH_EXPIRATION

    access_token_expiration = timezone.now() + jwt_settings.ACCESS_TOKEN_LIFETIME
    refresh_token_expiration = timezone.now() + jwt_settings.REFRESH_TOKEN_LIFETIME
    return_expiration_times = app_settings.JWT_AUTH_RETURN_EXPIRATION
    auth_httponly = app_settings.JWT_AUTH_HTTPONLY

    access_token, refresh_token = get_tokens_for_user(user)

    data = {
        "user": user,
        "access": access_token,
    }

    if not auth_httponly:
        data["refresh"] = refresh_token
    else:
        # Wasnt sure if the serializer needed this
        data["refresh"] = ""

    if return_expiration_times:
        data["access_expiration"] = access_token_expiration
        data["refresh_expiration"] = refresh_token_expiration

    serializer = serializer_class(
        instance=data,
        context=context,
    )

    return serializer.data


def verify(serializer, request, totp_purpose, login=True) -> Response:
    """
    If OTP is valid set user.is_active and the respective
    app_settings.VERIFICATION_METHOD of the user account to True.
    """
    otp = serializer.validated_data["otp"]
    success, totp = verify_otp(otp, totp_purpose)
    if not success:
        return Response({"detail": "Invalid OTP."}, status=status.HTTP_400_BAD_REQUEST)

    response_data = {}

    user_account = Account.objects.get(user=totp.user)
    if totp_purpose == TOTP.PURPOSE_ACCOUNT_VERIFICATION:
        user_account.is_verified = True
        response_data = {"detail": "Account verified successfully."}
    elif totp_purpose == TOTP.PURPOSE_EMAIL_VERIFICATION:
        user_account.email_verified = True
        response_data = {"detail": "Email verified successfully."}
    elif totp_purpose == TOTP.PURPOSE_PHONE_VERIFICATION:
        user_account.phone_verified = True
        response_data = {"detail": "Phone number verified successfully."}

    totp.user.is_active = True

    totp.user.save()
    user_account.save()

    response = Response(data=response_data, status=status.HTTP_200_OK)
    if login and app_settings.LOGIN_UPON_VERIFICATION:
        data = get_login_response_data(totp.user, {"request": request})
        response.data = {**response_data, **data}

        set_jwt_cookies(response, data["access"], data["refresh"])

    return response


class RegisterView(CreateAPIView):
    serializer_class = app_settings.REGISTER_SERIALIZER
    permission_classes = app_settings.REGISTER_PERMISSION_CLASSES

    def get_throttles(self):
        self.throttle_scope = get_throttle_scope("otp_auth_register")
        return super().get_throttles()

    @sensitive_post_parameters_m
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def get_response_data(self, user):
        if (
            app_settings.VERIFICATION_METHOD
            != app_settings.AccountVerificationMethod.NONE
        ):
            return {"detail": _("Verification OTP sent.")}

        return get_login_response_data(user, self.get_serializer_context())

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        user = self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        data = self.get_response_data(user)

        if data:
            response = Response(
                data,
                status=status.HTTP_201_CREATED,
                headers=headers,
            )
        else:
            response = Response(status=status.HTTP_204_NO_CONTENT, headers=headers)

        if app_settings.VERIFICATION_REQUIRED is False:
            user.is_active = True
            user.save()

        return response

    def perform_create(self, serializer):
        user = serializer.save(self.request)

        signal_kwargs = {}
        signals.user_signed_up.send(
            sender=user.__class__,
            request=self.request._request,
            user=user,
            **signal_kwargs,
        )

        # send OTP
        if (
            app_settings.VERIFICATION_METHOD
            == app_settings.AccountVerificationMethod.ACCOUNT
        ):
            purpose = TOTP.PURPOSE_ACCOUNT_VERIFICATION
        elif (
            app_settings.VERIFICATION_METHOD
            == app_settings.AccountVerificationMethod.EMAIL
        ):
            purpose = TOTP.PURPOSE_EMAIL_VERIFICATION
        elif (
            app_settings.VERIFICATION_METHOD
            == app_settings.AccountVerificationMethod.PHONE
        ):
            purpose = TOTP.PURPOSE_PHONE_VERIFICATION

        if (
            app_settings.VERIFICATION_METHOD
            != app_settings.AccountVerificationMethod.NONE
        ):
            totp = TOTP.objects.create(user=user, purpose=purpose)
            send_verification_otp(totp, signup=True)

        return user


class VerifyAccountView(views.APIView):
    permission_classes = (AllowAny,)
    allowed_methods = ("POST", "OPTIONS", "HEAD")

    def get_throttles(self):
        self.throttle_scope = get_throttle_scope("otp_auth_verify_account")
        return super().get_throttles()

    def get_serializer(self, *args, **kwargs):
        return app_settings.OTP_SERIALIZER(*args, **kwargs)

    def get(self, *args, **kwargs):
        raise MethodNotAllowed("GET")

    def post(self, request):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        return verify(serializer, request, TOTP.PURPOSE_ACCOUNT_VERIFICATION)


class VerifyEmailView(views.APIView):
    throttle_scope = "otp_auth_v_email"
    permission_classes = (AllowAny,)
    allowed_methods = ("POST", "OPTIONS", "HEAD")

    def get_throttles(self):
        self.throttle_scope = get_throttle_scope("otp_auth_verify_email")
        return super().get_throttles()

    def get_serializer(self, *args, **kwargs):
        return app_settings.OTP_SERIALIZER(*args, **kwargs)

    def get(self, *args, **kwargs):
        raise MethodNotAllowed("GET")

    def post(self, request):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        return verify(serializer, request, TOTP.PURPOSE_EMAIL_VERIFICATION)


class VerifyPhoneView(views.APIView):
    permission_classes = (AllowAny,)
    allowed_methods = ("POST", "OPTIONS", "HEAD")

    def get_throttles(self):
        self.throttle_scope = get_throttle_scope("otp_auth_verify_phone")
        return super().get_throttles()

    def get_serializer(self, *args, **kwargs):
        return app_settings.OTP_SERIALIZER(*args, **kwargs)

    def get(self, *args, **kwargs):
        raise MethodNotAllowed("GET")

    def post(self, request):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        return verify(serializer, request, TOTP.PURPOSE_PHONE_VERIFICATION)


class ResendOTPView(views.APIView):
    permission_classes = (AllowAny,)
    allowed_methods = ("POST", "OPTIONS", "HEAD")

    def get_throttles(self):
        self.throttle_scope = get_throttle_scope("otp_auth_otp_resend")
        return super().get_throttles()

    def get_serializer(self, *args, **kwargs):
        serializer = app_settings.RESEND_OTP_SERIALIZER
        return serializer(*args, **kwargs)

    def get(self, *args, **kwargs):
        raise MethodNotAllowed("GET")

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        data = serializer.validated_data

        if data["purpose"] == TOTP.PURPOSE_ACCOUNT_VERIFICATION:
            email, phone = data.get("email"), data.get("phone")
            filter = {app_settings.USER_MODEL_EMAIL_FIELD: email} if email else {}
            if phone:
                filter[app_settings.USER_MODEL_PHONE_FIELD] = phone

        elif data["purpose"] == TOTP.PURPOSE_EMAIL_VERIFICATION:
            filter = {app_settings.USER_MODEL_EMAIL_FIELD: data["email"]}
        elif data["purpose"] == TOTP.PURPOSE_PHONE_VERIFICATION:
            filter = {app_settings.USER_MODEL_PHONE_FIELD: data["phone"]}
        elif data["purpose"] == TOTP.PURPOSE_PASSWORD_RESET:
            email, phone = data.get("email"), data.get("phone")
            filter = {app_settings.USER_MODEL_EMAIL_FIELD: email} if email else {}
            if phone:
                filter[app_settings.USER_MODEL_PHONE_FIELD] = phone

        user = UserModel.objects.filter(**filter).first()
        if not user:
            return Response(
                {"detail": "Incorrect email or phone."},
                status=status.HTTP_400_BAD_REQUEST,
            )

        purpose = data["purpose"]
        account = Account.objects.filter(user=user).first()

        if purpose == TOTP.PURPOSE_ACCOUNT_VERIFICATION and account.is_verified:
            msg = "Account already verified."
            return Response({"detail": msg}, status=status.HTTP_400_BAD_REQUEST)
        elif purpose == TOTP.PURPOSE_EMAIL_VERIFICATION and account.email_verified:
            msg = "Email address already verified."
            return Response({"detail": msg}, status=status.HTTP_400_BAD_REQUEST)
        elif purpose == TOTP.PURPOSE_PHONE_VERIFICATION and account.phone_verified:
            msg = "Phone number already verified."
            return Response({"detail": msg}, status=status.HTTP_400_BAD_REQUEST)

        # invalidate existing otp
        old_totp = TOTP.objects.filter(user=user, purpose=data["purpose"]).first()
        if old_totp and old_totp.is_valid:
            old_totp.is_valid = False
            old_totp.save()

        new_totp = TOTP.objects.create(user=user, purpose=data["purpose"])
        send_verification_otp(new_totp, request=request)

        return Response({"detail": "ok"}, status=status.HTTP_200_OK)


class LoginView(GenericAPIView):
    permission_classes = (AllowAny,)
    serializer_class = app_settings.LOGIN_SERIALIZER

    def get_throttles(self):
        self.throttle_scope = get_throttle_scope("otp_auth_login")
        return super().get_throttles()

    user = None
    access_token = None
    token = None

    @sensitive_post_parameters_m
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.request = request
        self.serializer = self.get_serializer(data=self.request.data)
        self.serializer.is_valid(raise_exception=True)

        user = self.serializer.validated_data["user"]
        data = get_login_response_data(user, {"request": request})

        response = Response(data, status=status.HTTP_200_OK)
        set_jwt_cookies(response, data["access"], data["refresh"])

        return response


class LogoutView(GenericAPIView):
    """
    Delete the Token object assigned to the current User object.
    Accepts/Returns nothing.
    """

    permission_classes = [IsAuthenticated]
    serializer_class = app_settings.LOGOUT_SERIALIZER

    def get_throttles(self):
        self.throttle_scope = get_throttle_scope("otp_auth_logout")
        return super().get_throttles()

    def post(self, request, *args, **kwargs):
        self.request = request
        self.serializer = self.get_serializer(data=self.request.data)
        self.serializer.is_valid(raise_exception=True)

        return self.logout(request)

    def logout(self, request):
        response = Response(
            {"detail": _("Successfully logged out.")},
            status=status.HTTP_200_OK,
        )

        cookie_name = app_settings.JWT_AUTH_COOKIE
        unset_jwt_cookies(response)

        if "rest_framework_simplejwt.token_blacklist" in settings.INSTALLED_APPS:
            # add refresh token to blacklist
            try:
                token = RefreshToken(request.data["refresh"])
                token.blacklist()
            except KeyError:
                response.data = {
                    "detail": _("Refresh token was not included in request data.")
                }
                response.status_code = status.HTTP_401_UNAUTHORIZED
            except (TokenError, AttributeError, TypeError) as error:
                if hasattr(error, "args"):
                    if (
                        "Token is blacklisted" in error.args
                        or "Token is invalid or expired" in error.args
                    ):
                        response.data = {"detail": _(error.args[0])}
                        response.status_code = status.HTTP_401_UNAUTHORIZED
                    else:
                        response.data = {"detail": _("An error has occurred.")}
                        response.status_code = status.HTTP_500_INTERNAL_SERVER_ERROR

                else:
                    response.data = {"detail": _("An error has occurred.")}
                    response.status_code = status.HTTP_500_INTERNAL_SERVER_ERROR

        elif not cookie_name:
            message = _(
                "Neither cookies or blacklist are enabled, so the token "
                "has not been deleted server side. Please make sure the token is deleted client side.",
            )
            response.data = {"detail": message}
            response.status_code = status.HTTP_200_OK

        return response


class ResetPasswordView(GenericAPIView):
    permission_classes = (AllowAny,)
    serializer_class = app_settings.PASSWORD_RESET_SERIALIZER

    def get_throttles(self):
        self.throttle_scope = get_throttle_scope("otp_auth_password_reset")
        return super().get_throttles()

    def post(self, request, *args, **kwargs):
        self.request = request
        self.serializer = self.get_serializer(data=self.request.data)
        self.serializer.is_valid(raise_exception=True)

        user = self.serializer.validated_data["user"]
        if user:
            totp = TOTP.objects.create(user=user, purpose=TOTP.PURPOSE_PASSWORD_RESET)
            send_verification_otp(totp, request)

        return Response(
            {"detail": _("Verification OTP sent.")}, status=status.HTTP_200_OK
        )


class PasswordResetConfirmView(GenericAPIView):
    serializer_class = app_settings.PASSWORD_RESET_CONFIRM_SERIALIZER
    permission_classes = (AllowAny,)

    def get_throttles(self):
        self.throttle_scope = get_throttle_scope("otp_auth_password_reset_confirm")
        return super().get_throttles()

    @sensitive_post_parameters_m
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response(
            {"detail": _("Password has been reset with the new password.")},
        )


class PasswordChangeView(GenericAPIView):
    serializer_class = app_settings.PASSWORD_CHANGE_SERIALIZER
    permission_classes = (IsAuthenticated,)

    def get_throttles(self):
        self.throttle_scope = get_throttle_scope("otp_auth_password_change")
        return super().get_throttles()

    @sensitive_post_parameters_m
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        resp_detail = {"detail": "New password has been saved."}

        if app_settings.LOGOUT_ON_PASSWORD_CHANGE:
            logout_view = LogoutView()
            response = logout_view.logout(request)

            resp_detail["logout_detail"] = response.data["detail"]

        return Response(resp_detail)


class UserDetailsView(RetrieveUpdateAPIView):
    """
    Reads and updates UserModel fields
    Accepts GET, PUT, PATCH methods.

    Default accepted fields: username, first_name, last_name
    Default display fields: pk, username, email, first_name, last_name
    Read-only fields: pk, email

    Returns UserModel fields.
    """

    serializer_class = app_settings.USER_DETAILS_SERIALIZER
    permission_classes = (IsAuthenticated,)

    def get_object(self):
        return self.request.user

    def get_queryset(self):
        """
        Adding this method since it is sometimes called when using
        django-rest-swagger
        """
        return get_user_model().objects.none()


class InvokeChangeEmailOTPView(GenericAPIView):
    serializer_class = app_settings.CHANGE_EMAIL_SERIALIZER
    permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        email = serializer.validated_data["new_email"]
        user = request.user

        totp = TOTP.objects.create(user=user, purpose=TOTP.PURPOSE_EMAIL_VERIFICATION)
        TOTPMetadata.objects.create(totp=totp, new_email=email)

        adapter.send_otp_to_email(totp, email)

        return Response({"detail": _("Verification OTP sent.")})


class InvokeChangePhoneOTPView(GenericAPIView):
    serializer_class = app_settings.CHANGE_PHONE_SERIALIZER
    permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        phone = serializer.validated_data["new_phone"]
        user = request.user

        totp = TOTP.objects.create(user=user, purpose=TOTP.PURPOSE_PHONE_VERIFICATION)
        TOTPMetadata.objects.create(totp=totp, new_phone=phone)

        adapter.send_otp_to_phone(totp, phone)

        return Response({"detail": _("Verification OTP sent.")})


class ChangeEmailConfrimationView(GenericAPIView):
    serializer_class = app_settings.CHANGE_EMAIL_CONFIRM_SERIALIZER
    permission_classes = (IsAuthenticated,)

    def post(self, request):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        otp = serializer.validated_data["otp"]
        success, totp = verify_otp(otp, TOTP.PURPOSE_EMAIL_VERIFICATION)
        if not success or totp.user != request.user:
            return Response(
                {"detail": "Invalid OTP."}, status=status.HTTP_400_BAD_REQUEST
            )

        totp_metadata = TOTPMetadata.objects.filter(totp=totp).first()
        if not totp_metadata or not totp_metadata.new_email:
            return Response(
                {"detail": "Invalid OTP."}, status=status.HTTP_400_BAD_REQUEST
            )

        if app_settings.UNIQUE_EMAIL:
            # Ensure the new email address isn't used by another user
            email_field = app_settings.USER_MODEL_EMAIL_FIELD
            if UserModel.objects.filter(
                **{email_field: totp_metadata.new_email}
            ).exists():
                return Response(
                    {"detail": "Email address already exists."},
                    status=status.HTTP_400_BAD_REQUEST,
                )

        user_account = Account.objects.filter(user=totp.user).first()
        if totp.user and not user_account:
            # Create account for user. The user may have logged in via some social auth.
            user_account = Account.objects.create(user=totp.user)

        if not user_account.email_verified:
            user_account.email_verified = True

        totp.user.is_active = True
        new_email = totp_metadata.new_email
        setattr(totp.user, app_settings.USER_MODEL_EMAIL_FIELD, new_email)

        totp.user.save()
        user_account.save()

        return Response(
            {"detail": "Email address updated successfully."}, status=status.HTTP_200_OK
        )


class ChangePhoneConfirmationView(GenericAPIView):
    serializer_class = app_settings.CHANGE_PHONE_CONFIRM_SERIALIZER
    permission_classes = (IsAuthenticated,)

    def post(self, request):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        otp = serializer.validated_data["otp"]
        success, totp = verify_otp(otp, TOTP.PURPOSE_PHONE_VERIFICATION)
        if not success or totp.user != request.user:
            return Response(
                {"detail": "Invalid OTP."}, status=status.HTTP_400_BAD_REQUEST
            )

        totp_metadata = TOTPMetadata.objects.filter(totp=totp).first()
        if not totp_metadata or not totp_metadata.new_phone:
            return Response(
                {"detail": "Invalid OTP."}, status=status.HTTP_400_BAD_REQUEST
            )

        if app_settings.UNIQUE_PHONE:
            # Ensure the new phone number isn't used by another user
            phone_field = app_settings.USER_MODEL_PHONE_FIELD
            if UserModel.objects.filter(
                **{phone_field: totp_metadata.new_phone}
            ).exists():
                return Response(
                    {"detail": "Phone number already exists."},
                    status=status.HTTP_400_BAD_REQUEST,
                )

        user_account = Account.objects.filter(user=totp.user).first()
        if totp.user and not user_account:
            # Create account for user. The user may have logged in via some social auth.
            user_account = Account.objects.create(user=totp.user)

        if not user_account.phone_verified:
            user_account.phone_verified = True

        totp.user.is_active = True
        new_phone = totp_metadata.new_phone
        setattr(totp.user, app_settings.USER_MODEL_PHONE_FIELD, new_phone)

        totp.user.save()
        user_account.save()
        return Response(
            {"detail": "Phone number updated successfully."}, status=status.HTTP_200_OK
        )
